import { validateAndParseAddress } from './utils';
import { CurrencyAmount, Price, ChainId, Token } from '@sushiswap/sdk'
import { getSignatureWithProvider, getSignature, Message, getTypedData, getTypeHash } from './entities/eip712'
import { LAMBDA_URL } from './constants';
import { ethers } from 'ethers';
import axios from "axios";

export interface ILimitOrderData {
    maker: string;
    tokenIn: string;
    tokenOut: string;
    tokenInDecimals: number;
    tokenOutDecimals: number;
    tokenInSymbol: string;
    tokenOutSymbol: string;
    amountIn: string;
    amountOut: string;
    recipient: string;
    startTime: string;
    endTime: string;
    stopPrice?: string;
    oracleAddress?: string;
    oracleData?: string;
    v: number;
    r: string;
    s: string;
    chainId: ChainId;
    orderTypeHash?: string;
}

export class LimitOrder {

    public readonly maker: string;
    public readonly amountIn: CurrencyAmount<Token>;
    public readonly amountOut: CurrencyAmount<Token>;
    public readonly recipient: string;
    public readonly startTime: string;
    public readonly endTime: string;
    public readonly stopPrice: string;
    public readonly oracleAddress: string;
    public readonly oracleData: string;
    public v: number;
    public r: string;
    public s: string;

    static getLimitOrder(data: ILimitOrderData): LimitOrder {
        return new LimitOrder(
            data.maker,
            CurrencyAmount.fromRawAmount(new Token(data.chainId, data.tokenIn, data.tokenInDecimals, data.tokenInSymbol), data.amountIn),
            CurrencyAmount.fromRawAmount(new Token(data.chainId, data.tokenOut, data.tokenOutDecimals, data.tokenOutSymbol), data.amountOut),
            data.recipient,
            data.startTime,
            data.endTime,
            data.stopPrice,
            data.oracleAddress,
            data.oracleData,
            data.v,
            data.r,
            data.s
        );
    }

    constructor(
        maker: string,
        amountIn: CurrencyAmount<Token>,
        amountOut: CurrencyAmount<Token>,
        recipient: string,
        startTime: string,
        endTime: string,
        stopPrice = "0",
        oracleAddress = "0x0000000000000000000000000000000000000000",
        oracleData = "0x00000000000000000000000000000000000000000000000000000000000000",
        v = 0,
        r = "",
        s = ""
    ) { 
        this.maker = validateAndParseAddress(maker);
        this.amountIn = amountIn;
        this.amountOut = amountOut;
        this.recipient = validateAndParseAddress(recipient);
        this.startTime = startTime;
        this.endTime = endTime;
        this.stopPrice = stopPrice;
        this.oracleAddress = validateAndParseAddress(oracleAddress);
        this.oracleData = oracleData;
        this.v = v;
        this.r = r;
        this.s = s;
    }

    get amountInRaw(): string {
        return this.amountIn.quotient.toString();
    }

    get amountOutRaw(): string {
        return this.amountOut.quotient.toString();
    }

    get tokenInAddress(): string {
        return this.amountIn.currency.address;
    }

    get tokenOutAddress(): string {
        return this.amountOut.currency.address;
    }

    get tokenInDecimals(): number {
        return this.amountIn.currency.decimals;
    }

    get tokenOutDecimals(): number {
        return this.amountOut.currency.decimals;
    }

    get tokenInSymbol(): string {
        return this.amountIn.currency.symbol || '';
    }

    get tokenOutSymbol(): string {
        return this.amountOut.currency.symbol || '';
    }

    get chainId(): ChainId {
        return this.amountIn.currency.chainId;
    }

    usePrice(price: Price<Token, Token>): LimitOrder {
        return new LimitOrder(
            this.maker,
            this.amountIn,
            CurrencyAmount.fromRawAmount(this.amountOut.currency, price.quote(this.amountIn).quotient.toString()),
            this.recipient,
            this.startTime,
            this.endTime,
            this.stopPrice,
            this.oracleAddress,
            this.oracleData
        );
    }

    signdOrderWithPrivatekey(chainId: ChainId, privateKey: string) {
        let order : Message = {
            maker: this.maker,
            tokenIn: this.tokenInAddress,
            tokenOut: this.tokenOutAddress,
            amountIn: this.amountInRaw,
            amountOut: this.amountOutRaw,
            recipient: this.recipient,
            startTime: this.startTime,
            endTime: this.endTime,
            stopPrice: this.stopPrice,
            oracleAddress : this.oracleAddress,
            oracleData: ethers.utils.keccak256(this.oracleData)
        }

        const { v, r, s } = getSignature(order, chainId, privateKey);

        this.v = v;
        this.r = r;
        this.s = s;

        return { v, r, s }
    }

    async signOrderWithProvider(chainId: ChainId, provider: ethers.providers.Web3Provider) {
        let order: Message = {
            maker: this.maker,
            tokenIn: this.tokenInAddress,
            tokenOut: this.tokenOutAddress,
            amountIn: this.amountInRaw,
            amountOut: this.amountOutRaw,
            recipient: this.recipient,
            startTime: this.startTime,
            endTime: this.endTime,
            stopPrice: this.stopPrice,
            oracleAddress: this.oracleAddress,
            oracleData: ethers.utils.keccak256(this.oracleData)
        }

        const { v, r, s } = await getSignatureWithProvider(order, chainId, provider);

        this.v = v;
        this.r = r;
        this.s = s;

        return { v, r, s }
    }

    getTypedData() {

        let order : Message = {
            maker: this.maker,
            tokenIn: this.tokenInAddress,
            tokenOut: this.tokenOutAddress,
            amountIn: this.amountInRaw,
            amountOut: this.amountOutRaw,
            recipient : this.recipient,
            startTime: this.startTime,
            endTime: this.endTime,
            stopPrice: this.stopPrice,
            oracleAddress : this.oracleAddress,
            oracleData: ethers.utils.keccak256(this.oracleData)
        }

        return getTypedData(order, this.chainId);
    }

    getTypeHash() {
        let typedData = this.getTypedData();
        let digest = getTypeHash(typedData);
        return digest;
    }

    async send() {
        const resp = await axios.post(`${LAMBDA_URL}/orders/create`, {
            maker: this.maker,
            tokenIn: this.tokenInAddress,
            tokenOut: this.tokenOutAddress,
            tokenInDecimals: this.tokenInDecimals,
            tokenOutDecimals: this.tokenOutDecimals,
            tokenInSymbol: this.tokenInSymbol,
            tokenOutSymbol: this.tokenOutSymbol,
            amountIn: this.amountInRaw,
            amountOut: this.amountOutRaw,
            recipient: this.recipient,
            startTime: this.startTime,
            endTime: this.endTime,
            stopPrice: this.stopPrice,
            oracleAddress: this.oracleAddress,
            oracleData: this.oracleData,
            v: this.v,
            r: this.r,
            s: this.s,
            chainId: this.amountIn.currency.chainId
        });

        return resp.data;
    }
}
